// MIT License
//
// Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//

#pragma once

#include <cstdint>
#include <iomanip>
#include <sstream>

namespace rocprofiler
{
namespace sdk
{
namespace utility
{
template <typename Tp>
auto
as_hex(Tp val, size_t width = 0)
{
    uintptr_t _uintp_val = 0;
    if constexpr(std::is_pointer<Tp>::value)
        _uintp_val = reinterpret_cast<uintptr_t>(val);
    else
        _uintp_val = val;

    auto ss = std::stringstream{};
    ss << "0x" << std::hex << std::setfill('0') << std::setw(width) << _uintp_val;
    return ss.str();
}
}  // namespace utility
}  // namespace sdk
}  // namespace rocprofiler
